@@ -14,15 +14,9 @@ See the License for the specific language governing permissions and
 limitations under the License.
 """
 from pynetbox.core.query import Request, RequestError
-from pynetbox.core.response import Record
+from pynetbox.core.response import Record, RecordSet
 
-RESERVED_KWARGS = ("id", "pk", "limit", "offset")
-
-
-def response_loader(req, return_obj, endpoint):
-    if isinstance(req, list):
-        return [return_obj(i, endpoint.api, endpoint) for i in req]
-    return return_obj(req, endpoint.api, endpoint)
+RESERVED_KWARGS = ()
 
 
 class Endpoint(object):
@@ -54,7 +48,9 @@ class Endpoint(object):
         self.token = api.token
         self.session_key = api.session_key
         self.url = "{base_url}/{app}/{endpoint}".format(
-            base_url=self.base_url, app=app.name, endpoint=self.name,
+            base_url=self.base_url,
+            app=app.name,
+            endpoint=self.name,
         )
         self._choices = None
 
@@ -77,28 +73,41 @@ class Endpoint(object):
             ret = Record
         return ret
 
-    def all(self):
+    def all(self, limit=0, offset=None):
         """Queries the 'ListView' of a given endpoint.
 
         Returns all objects from an endpoint.
 
-        :Returns: List of :py:class:`.Record` objects.
+        :arg int,optional limit: Overrides the max page size on
+            paginated returns.
+        :arg int,optional offset: Overrides the offset on paginated returns.
+
+        :Returns: A :py:class:`.RecordSet` object.
 
         :Examples:
 
-        >>> nb.dcim.devices.all()
-        [test1-a3-oobsw2, test1-a3-oobsw3, test1-a3-oobsw4]
+        >>> devices = nb.dcim.devices.all()
+        >>> for device in devices:
+        ...     print(device.name)
+        ...
+        test1-leaf1
+        test1-leaf2
+        test1-leaf3
         >>>
         """
+        if limit == 0 and offset is not None:
+            raise ValueError("offset requires a positive limit value")
         req = Request(
             base="{}/".format(self.url),
             token=self.token,
             session_key=self.session_key,
             http_session=self.api.http_session,
             threading=self.api.threading,
+            limit=limit,
+            offset=offset,
         )
 
-        return response_loader(req.get(), self.return_obj, self)
+        return RecordSet(self, req)
 
     def get(self, *args, **kwargs):
         r"""Queries the DetailsView of a given endpoint.
@@ -135,17 +144,19 @@ class Endpoint(object):
             key = None
 
         if not key:
-            filter_lookup = self.filter(**kwargs)
-            if filter_lookup:
-                if len(filter_lookup) > 1:
-                    raise ValueError(
-                        "get() returned more than one result. "
-                        "Check that the kwarg(s) passed are valid for this "
-                        "endpoint or use filter() or all() instead."
-                    )
-                else:
-                    return filter_lookup[0]
-            return None
+            resp = self.filter(**kwargs)
+            ret = next(resp, None)
+            if not ret:
+                return ret
+            try:
+                next(resp)
+                raise ValueError(
+                    "get() returned more than one result. "
+                    "Check that the kwarg(s) passed are valid for this "
+                    "endpoint or use filter() or all() instead."
+                )
+            except StopIteration:
+                return ret
 
         req = Request(
             key=key,
@@ -154,17 +165,14 @@ class Endpoint(object):
             session_key=self.session_key,
             http_session=self.api.http_session,
         )
-
         try:
-            resp = req.get()
+            return next(RecordSet(self, req), None)
         except RequestError as e:
             if e.req.status_code == 404:
                 return None
             else:
                 raise e
 
-        return response_loader(resp, self.return_obj, self)
-
     def filter(self, *args, **kwargs):
         r"""Queries the 'ListView' of a given endpoint.
 
@@ -176,48 +184,69 @@ class Endpoint(object):
             accepted on given endpoint.
         :arg str,optional \**kwargs: Any search argument the
             endpoint accepts can be added as a keyword arg.
+        :arg int,optional limit: Overrides the max page size on
+            paginated returns.
+        :arg int,optional offset: Overrides the offset on paginated returns.
 
-        :Returns: A list of :py:class:`.Record` objects.
+        :Returns: A :py:class:`.RecordSet` object.
 
         :Examples:
 
         To return a list of objects matching a named argument filter.
 
-        >>> nb.dcim.devices.filter(role='leaf-switch')
-        [test1-a3-tor1b, test1-a3-tor1c, test1-a3-tor1d, test1-a3-tor2a]
+        >>> devices = nb.dcim.devices.filter(role='leaf-switch')
+        >>> for device in devices:
+        ...     print(device.name)
+        ...
+        test1-leaf1
+        test1-leaf2
+        test1-leaf3
         >>>
 
         Using a freeform query along with a named argument.
 
-        >>> nb.dcim.devices.filter('a3', role='leaf-switch')
-        [test1-a3-tor1b, test1-a3-tor1c, test1-a3-tor1d, test1-a3-tor2a]
+        >>> devices = nb.dcim.devices.filter('a3', role='leaf-switch')
+        >>> for device in devices:
+        ...     print(device.name)
+        ...
+        test1-a3-leaf1
+        test1-a3-leaf2
         >>>
 
         Chaining multiple named arguments.
 
-        >>> nb.dcim.devices.filter(role='leaf-switch', status=True)
-        [test1-leaf2]
+        >>> devices = nb.dcim.devices.filter(role='leaf-switch', status=True)
+        >>> for device in devices:
+        ...     print(device.name)
+        ...
+        test1-leaf2
         >>>
 
         Passing a list as a named argument adds multiple filters of the
         same value.
 
-        >>> nb.dcim.devices.filter(role=['leaf-switch', 'spine-switch'])
-        [test1-a3-spine1, test1-a3-spine2, test1-a3-leaf1]
+        >>> device = nb.dcim.devices.filter(role=['leaf-switch', 'spine-switch'])
+        >>> for device in devices:
+        ...     print(device.name)
+        ...
+        test1-a3-spine1
+        test1-a3-spine2
+        test1-a3-leaf1
         >>>
         """
 
         if args:
             kwargs.update({"q": args[0]})
 
-        if not kwargs:
-            raise ValueError("filter must be passed kwargs. Perhaps use all() instead.")
         if any(i in RESERVED_KWARGS for i in kwargs):
             raise ValueError(
-                "A reserved {} kwarg was passed. Please remove it "
-                "try again.".format(RESERVED_KWARGS)
+                "A reserved kwarg was passed ({}). Please remove it "
+                "and try again.".format(RESERVED_KWARGS)
             )
-
+        limit = kwargs.pop("limit") if "limit" in kwargs else 0
+        offset = kwargs.pop("offset") if "offset" in kwargs else None
+        if limit == 0 and offset is not None:
+            raise ValueError("offset requires a positive limit value")
         req = Request(
             filters=kwargs,
             base=self.url,
@@ -225,9 +254,11 @@ class Endpoint(object):
             session_key=self.session_key,
             http_session=self.api.http_session,
             threading=self.api.threading,
+            limit=limit,
+            offset=offset,
         )
 
-        return response_loader(req.get(), self.return_obj, self)
+        return RecordSet(self, req)
 
     def create(self, *args, **kwargs):
         r"""Creates an object on an endpoint.
@@ -251,10 +282,9 @@ class Endpoint(object):
 
         :Examples:
 
-        Creating an object on the `devices` endpoint you can lookup a
-        device_role's name with:
+        Creating an object on the `devices` endpoint:
 
-        >>> netbox.dcim.devices.create(
+        >>> device = netbox.dcim.devices.create(
         ...    name='test',
         ...    device_role=1,
         ... )
@@ -287,10 +317,135 @@ class Endpoint(object):
             http_session=self.api.http_session,
         ).post(args[0] if args else kwargs)
 
-        return response_loader(req, self.return_obj, self)
+        if isinstance(req, list):
+            return [self.return_obj(i, self.api, self) for i in req]
+        return self.return_obj(req, self.api, self)
+
+    def update(self, objects):
+        r"""Bulk updates existing objects on an endpoint.
+
+        Allows for bulk updating of existing objects on an endpoint.
+        Objects is a list whic contain either json/dicts or Record
+        derived objects, which contain the updates to apply.
+        If json/dicts are used, then the id of the object *must* be
+        included
+
+        :arg list objects: A list of dicts or Record.
+
+        :returns: True if the update succeeded
+
+        :Examples:
+
+        Updating objects on the `devices` endpoint:
+
+        >>> device = netbox.dcim.devices.update([
+        ...    {'id': 1, 'name': 'test'},
+        ...    {'id': 2, 'name': 'test2'},
+        ... ])
+        >>> True
+
+        Use bulk update by passing a list of Records:
+
+        >>> devices = nb.dcim.devices.all()
+        >>> for d in devices:
+        >>>     d.name = d.name+'-test'
+        >>> nb.dcim.devices.update(devices)
+        >>> True
+        """
+        series = []
+        if not isinstance(objects, list):
+            raise ValueError(
+                "Objects passed must be list[dict|Record] - was " + type(objects)
+            )
+        for o in objects:
+            if isinstance(o, Record):
+                data = o.updates()
+                if data:
+                    data["id"] = o.id
+                    series.append(data)
+            elif isinstance(o, dict):
+                if "id" not in o:
+                    raise ValueError("id is missing from object: " + str(o))
+                series.append(o)
+            else:
+                raise ValueError(
+                    "Object passed must be dict|Record - was " + type(objects)
+                )
+        req = Request(
+            base=self.url,
+            token=self.token,
+            session_key=self.session_key,
+            http_session=self.api.http_session,
+        ).patch(series)
+
+        if isinstance(req, list):
+            return [self.return_obj(i, self.api, self) for i in req]
+        return self.return_obj(req, self.api, self)
+
+    def delete(self, objects):
+        r"""Bulk deletes objects on an endpoint.
+
+        Allows for batch deletion of multiple objects from
+        a single endpoint
+
+        :arg list objects: A list of either ids or Records or
+            a single RecordSet to delete.
+        :returns: True if bulk DELETE operation was successful.
+
+        :Examples:
+
+        Deleting all `devices`:
+
+        >>> netbox.dcim.devices.delete(netbox.dcim.devices.all(0))
+        >>>
+
+        Use bulk deletion by passing a list of ids:
+
+        >>> netbox.dcim.devices.delete([2, 243, 431, 700])
+        >>>
+
+        Use bulk deletion to delete objects eg. when filtering
+        on a `custom_field`:
+        >>> netbox.dcim.devices.delete([
+        >>>         d for d in netbox.dcim.devices.all(0) \
+        >>>             if d.custom_fields.get('field', False)
+        >>>     ])
+        >>>
+        """
+        cleaned_ids = []
+        if not isinstance(objects, list) and not isinstance(objects, RecordSet):
+            raise ValueError(
+                "objects must be list[str|int|Record]"
+                "|RecordSet - was " + str(type(objects))
+            )
+        for o in objects:
+            if isinstance(o, int):
+                cleaned_ids.append(o)
+            elif isinstance(o, str) and o.isnumeric():
+                cleaned_ids.append(int(o))
+            elif isinstance(o, Record):
+                if not hasattr(o, "id"):
+                    raise ValueError(
+                        "Record from '"
+                        + o.url
+                        + "' does not have an id and cannot be bulk deleted"
+                    )
+                cleaned_ids.append(o.id)
+            else:
+                raise ValueError(
+                    "Invalid object in list of " "objects to delete: " + str(type(o))
+                )
+
+        req = Request(
+            base=self.url,
+            token=self.token,
+            session_key=self.session_key,
+            http_session=self.api.http_session,
+        )
+        return True if req.delete(data=[{"id": i} for i in cleaned_ids]) else False
 
     def choices(self):
-        """ Returns all choices from the endpoint.
+        """Returns all choices from the endpoint.
 
         The returned dict is also saved in the endpoint object (in
         ``_choices`` attribute) so that later calls will return the same data
@@ -392,7 +547,7 @@ class Endpoint(object):
 
 
 class DetailEndpoint(object):
-    """Enables read/write Operations on detail endpoints.
+    """Enables read/write operations on detail endpoints.
 
     Endpoints like ``available-ips`` that are detail routes off
     traditional endpoints are handled with this class.
@@ -419,13 +574,18 @@ class DetailEndpoint(object):
             E.g. ``.list(method='get_facts')`` would be converted to
             ``.../?method=get_facts``.
 
-        :returns: A dictionary or list of dictionaries retrieved from
-            NetBox.
+        :returns: A :py:class:`.Record` object or list of :py:class:`.Record` objects created
+            from data retrieved from NetBox.
         """
         req = Request(**self.request_kwargs).get(add_params=kwargs)
 
         if self.custom_return:
-            return response_loader(req, self.custom_return, self.parent_obj.endpoint)
+            return [
+                self.custom_return(
+                    i, self.parent_obj.endpoint.api, self.parent_obj.endpoint
+                )
+                for i in req
+            ]
         return req
 
     def create(self, data=None):
@@ -438,13 +598,23 @@ class DetailEndpoint(object):
             object. Defaults to empty dict which will create a single
             item with default values.
 
-        :returns: A dictionary or list of dictionaries its created in
-            NetBox.
+        :returns: A :py:class:`.Record` object or list of :py:class:`.Record` objects created
+            from data created in NetBox.
         """
         data = data or {}
         req = Request(**self.request_kwargs).post(data)
         if self.custom_return:
-            return response_loader(req, self.custom_return, self.parent_obj.endpoint)
+            if isinstance(req, list):
+                return [
+                    self.custom_return(
+                        req_item, self.parent_obj.endpoint.api, self.parent_obj.endpoint
+                    )
+                    for req_item in req
+                ]
+            else:
+                return self.custom_return(
+                    req, self.parent_obj.endpoint.api, self.parent_obj.endpoint
+                )
         return req
 
 
